
"""

"""


from __future__ import absolute_import, division, print_function

import numpy as np
import pandas as pd
import datetime
from dateutil.relativedelta import relativedelta
import uuid
from causaltest.bin.conf import ALLOWED_TARGET, ALLOWED_COUNTRY_CODES, \
                                ALLOWED_GEO_DIM, DATASET_INCREMENTALITY, \
                                FACT_INCREMENTALITY_TEST, QUERIES, DATE_NAME,\
                                QUERY_DIM_GEO, GEO_SPLIT # TODO: change name of queries to be consistent
from causaltest.misc import str_to_date
from causaltest.infer import Infer
from causaltest.plot import Plot
from causaltest.split import perform_stratified_random_split
from causaltest.map import Mapper
from causaltest.io import custom_write_gbq, get_bigquery_client, load_fact_test, load_fact_split


class BaseTest(Infer, Plot, Mapper):
    """

    """
    def __init__(self, test_id: str, start_date: datetime.date, end_date: datetime.date, target: str,
                 country_code: str, geo_dim: str, status_activated: bool, **kwargs):
        self.test_id = test_id
        self.test_name = test_name
        self.start_date = start_date
        self.end_date = end_date
        self.target = target
        self.country_code = country_code # TODO: faire la distinction entre le scope et la granularité
        self.geo_dim = geo_dim
        self.status_activated = status_activated
        self.treatment_geo_id = []

    def activate_test(self):
        self.status_activated = True
        self.write_fact_test()

    def deactivate_test(self):
        self.status_activated = False
        self.write_fact_test()

    def process_split(self):
        self.treatment_geo_id, self.control_geo_id, self.exclude_geo_id = \
                perform_stratified_random_split(self.fact_geo, self.dim_geo)

    def write_split(self, func_write_to_db=custom_write_gbq,
                    client=get_bigquery_client()):
        df = pd.concat((
            pd.DataFrame(data={
                "test_id": self.test_id,
                "geo_dim_key": self.geo_dim,
                "geo_dim_value": self.treatment_geo_id,
                "group": "treatment"
            }),
            pd.DataFrame(data={
                "test_id": self.test_id,
                "geo_dim_key": self.geo_dim,
                "geo_dim_value": self.control_geo_id,
                "group": "control"
            }),
            pd.DataFrame(data={
                "test_id": self.test_id,
                "geo_dim_key": self.geo_dim,
                "geo_dim_value": self.exclude_geo_id,
                "group": "excluded"
            }),
        ))

        # TODO replace based on test_id
        func_write_to_db(client=client, target_dataset_name=DATASET_INCREMENTALITY,
                         target_table_name=GEO_SPLIT, df=df, mode='append')

    # TODO: think conf should be loaded with io functions
    def write_fact_test(self, func_write_to_db=custom_write_gbq,
                        client=get_bigquery_client()):
        df = pd.DataFrame(data={
            "test_id": [self.test_id],
            "creation_datetime": [datetime.datetime.now().replace(microsecond=0)],
            "start_date": [self.start_date],
            "end_date": [self.end_date],
            "target": [self.target],
            "geo_dim": [self.geo_dim],
            "country_code": [self.country_code],
            "status_activated": [self.status_activated],
            "test_name": [self.test_name]
        }) # TODO: change schema of destination table
        func_write_to_db(client=client, target_dataset_name=DATASET_INCREMENTALITY,
                         target_table_name=FACT_INCREMENTALITY_TEST, df=df, mode='append') # TODO: check if append is the best (keeps history)


class CausalTest(BaseTest):
    """

    """
    def __init__(self, test_id: str, start_date: datetime.date, end_date: datetime.date, target: str,
                 country_code: str, geo_dim: str, status_activated: bool, **kwargs):
        checked_input = self._process_input_data(test_id, start_date, end_date,
                                                 target, country_code,
                                                 geo_dim, status_activated,
                                                 **kwargs)
        super(CausalTest, self).__init__(**checked_input)
        #self.load_data() # TODO: faire une classmethod qui permet la création depuis un test_id ou de nouveaux parameters

    def _process_input_data(self, test_id, start_date, end_date, target,
                            country_code, geo_dim, status_activated, **kwargs):
        """
        Checks and formats when appropriate the input data for running the Test.
        Performs assertions such as missing or invalid arguments.
        """
        input_args = locals().copy()
        none_args = [arg for arg, value in input_args.items() if value is None]
        if none_args:
            raise ValueError('{args} input cannot be empty'.format(
                             args=', '.join(none_args)))
        # processed_data = self._format_input_data(data)
        # pre_data, post_data = self._process_pre_post_data(processed_data, pre_period,
        #                                                   post_period)
        assert target in ALLOWED_TARGET
        assert start_date < end_date
        assert country_code in ALLOWED_COUNTRY_CODES
        assert geo_dim in ALLOWED_GEO_DIM

        return {
            'test_id': test_id,
            'start_date': start_date,
            'end_date': end_date,
            'target': target,
            'country_code': country_code,
            'geo_dim': geo_dim,
            'status_activated': status_activated # TODO: asset it is a bool
        }

    @classmethod
    def get_from_id(cls, test_id, func_load_fact_test=load_fact_test):
        params = func_load_fact_test(test_id)
        params = params.sort_values('creation_datetime', ascending=False)\
                       .to_dict('records')[0]
        return CausalTest(params['test_id'], params['start_date'],
                          params['end_date'], params['target'],
                          params['country_code'], params['geo_dim'],
                          params['status_activated'])

    @classmethod
    def create_from_json(cls, json_obj):

        start_date = str_to_date(json_obj['start_date'])
        if 'end_date' not in json_obj.keys():
            end_date = start_date + relativedelta(months=+1)
        elif json_obj['end_date'] is None:
            end_date = start_date + relativedelta(months=+1)
        else:
            end_date = str_to_date(json_obj['end_date'])

        if 'target' not in json_obj.keys():
            target = ALLOWED_TARGET[0]
        else:
            target = json_obj['target']

        if 'country_code' not in json_obj.keys():
            country_code = ALLOWED_COUNTRY_CODES[0]
        else:
            country_code = json_obj['country_code']

        if 'geo_dim' not in json_obj.keys():
            geo_dim = ALLOWED_GEO_DIM[0]
        else:
            geo_dim = json_obj['geo_dim']

        if 'test_id' in json_obj.keys():
            test_id = json_obj['test_id']
        else:
            test_id = str(uuid.uuid4())

        tst = CausalTest(test_id=test_id, start_date=start_date, end_date=end_date,
                         target=target, country_code=country_code, geo_dim=geo_dim, status_activated=False)
        tst.write_fact_test()

        return tst

    @classmethod
    def get_from_pickle(cls, pickle_file):
        pass # TODO: see if necessary

    def map_split(self, to_html=False): # TODO: i think it can be exported to class Mapper
        """
        Create folium map invoking `Mapper` class.
        """
        self._generate_map()
        if to_html:
            return self.network_map_html
        return self.network_map

    def infer_impact(self):
        """
        Infer impact invoking `Inferences`class.
        """
        self._infer_causal_impact()
        return self.ci.summary_data
