"""
Plots the analysis obtained in causal impact algorithm.
"""


from __future__ import absolute_import, division, print_function

import io
import base64
import pandas as pd
import numpy as np
#import matplotlib.pyplot as plt
from matplotlib.backends.backend_agg import FigureCanvasAgg as FigureCanvas
from matplotlib.figure import Figure
import matplotlib.ticker

def myLogFormat(y,pos):
    # Find the number of decimal places required
    decimalplaces = int(np.maximum(-np.log10(y),0))     # =0 for numbers >=1
    # Insert that number into a format string
    formatstring = '{{:.{:1d}f}}'.format(decimalplaces)
    # Return the formatted tick label
    return formatstring.format(y)


class Plot(object):
    """Takes all the vectors and final analysis performed in the post-period inference
    to plot final graphics.
    """
    def plot_impact(self, panels=['original', 'pointwise', 'cumulative'],
                    figsize=(15, 12), to_html=False):
        """Plots inferences results related to causal impact analysis.

        Args
        ----
          panels: list.
            Indicates which plot should be considered in the graphics.
          figsize: tuple.
            Changes the size of the graphics plotted.

        Raises
        ------
          RuntimeError: if inferences were not computed yet.
        """
        # plt = self.ci._get_plotter()
        fig = Figure(figsize=figsize)
        # fig = plt.figure(figsize=figsize)
        if self.ci.summary_data is None:
            raise RuntimeError('Please first run inferences before plotting results')

        valid_panels = ['original', 'pointwise', 'cumulative']
        for panel in panels:
            if panel not in valid_panels:
                raise ValueError(
                    '"{}" is not a valid panel. Valid panels are: {}.'.format(
                        panel, ', '.join(['"{}"'.format(e) for e in valid_panels])
                    )
                )

        # First points can be noisy due approximation techniques used in the likelihood
        # optimizaion process. We remove those points from the plots.
        llb = self.ci.trained_model.filter_results.loglikelihood_burn
        inferences = self.ci.inferences.iloc[llb:]

        intervention_idx = inferences.index.get_loc(self.ci.post_period[0])
        n_panels = len(panels)
        ax = fig.add_subplot(n_panels, 1, 1)
        idx = 1

        if 'original' in panels:
            ax.plot(pd.concat([self.ci.pre_data.iloc[llb:, 0], self.ci.post_data.iloc[:, 0]]), # TODO: get rid of self.data_ci already in self.ci
                    'k', label='y')
            ax.plot(inferences['preds'], 'b--', label='Predicted')
            if panels == ['original']:  # panels only on original for pre test analysis
                ax.axvline(inferences.index[intervention_idx - 1], c='k', linestyle='--')
            ax.fill_between(
                self.ci.pre_data.index[llb:].union(self.ci.post_data.index),
                inferences['preds_lower'],
                inferences['preds_upper'],
                facecolor='blue',
                interpolate=True,
                alpha=0.25
            )
            ax.grid(True, linestyle='--')
            ax.legend()
            # if idx != n_panels:
            #     fig.setp(ax.get_xticklabels(), visible=False)
            idx += 1

        if 'pointwise' in panels:
            ax = fig.add_subplot(n_panels, 1, idx, sharex=ax)
            ax.plot(inferences['point_effects'], 'b--', label='Point Effects')
            ax.axvline(inferences.index[intervention_idx - 1], c='k', linestyle='--')
            ax.fill_between(
                inferences['point_effects'].index,
                inferences['point_effects_lower'],
                inferences['point_effects_upper'],
                facecolor='blue',
                interpolate=True,
                alpha=0.25
            )
            ax.axhline(y=0, color='k', linestyle='--')
            ax.grid(True, linestyle='--')
            ax.legend()
            # if idx != n_panels:
            #     fig.setp(ax.get_xticklabels(), visible=False)
            idx += 1

        if 'cumulative' in panels:
            ax = fig.add_subplot(n_panels, 1, idx, sharex=ax)
            ax.plot(inferences['post_cum_effects'], 'b--',
                    label='Cumulative Effect')
            ax.axvline(inferences.index[intervention_idx - 1], c='k', linestyle='--')
            ax.fill_between(
                inferences['post_cum_effects'].index,
                inferences['post_cum_effects_lower'],
                inferences['post_cum_effects_upper'],
                facecolor='blue',
                interpolate=True,
                alpha=0.25
            )
            ax.grid(True, linestyle='--')
            ax.axhline(y=0, color='k', linestyle='--')
            ax.legend()

        # Alert if points were removed due to loglikelihood burning data
        if llb > 0:
            text = ('Note: The first {} observations were removed due to approximate '
                    'diffuse initialization.'.format(llb))
            fig.text(0.1, 0.01, text, fontsize='large')

        if to_html:
            # Convert plot to PNG image
            pngImage = io.BytesIO()
            FigureCanvas(fig).print_png(pngImage)

            # Encode PNG image to base64 string
            pngImageB64String = "data:image/png;base64,"
            pngImageB64String += base64.b64encode(pngImage.getvalue())\
                                       .decode('utf8')

            return pngImageB64String


        return fig

    def plot_test_period(self, figsize=(15, 12), to_html=False):
        w = (self.ci.summary_data.loc['abs_effect_upper', 'average']\
             -self.ci.summary_data.loc['abs_effect_lower', 'average'])\
             /2/self.ci.summary_data.loc['actual', 'average'] # TODO: should be done in infer

        x = np.arange(1, 60)
        y = [(1+w)**(1/j)-1 for j in x]

        fig = Figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.plot(x+7, y) # at least one week of analysis
        ax.set_xticks(np.arange(5, 60, 5))
        ax.get_xaxis().set_major_formatter(matplotlib.ticker.ScalarFormatter())
        ax.set_yscale('log')
        ax.set_yticks([0.005, 0.01, 0.02, 0.05, 0.1, 0.2])
        ax.get_yaxis().set_major_formatter(matplotlib.ticker.ScalarFormatter())
        ax.grid(True, linestyle='--')
        ax.set_title('Number of days required to conclude on forecasted impact')

        if to_html:
            # Convert plot to PNG image
            pngImage = io.BytesIO()
            FigureCanvas(fig).print_png(pngImage)

            # Encode PNG image to base64 string
            pngImageB64String = "data:image/png;base64,"
            pngImageB64String += base64.b64encode(pngImage.getvalue())\
                                       .decode('utf8')

            return pngImageB64String

        return fig
